/*
 * Copyright (c) 2003-2005 Sendmail, Inc. and its suppliers.
 *      All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 *
 *	$Id: tls.h,v 1.70 2014/03/05 00:21:16 ca Exp $
 */

#ifndef SM_TLS_H
#define SM_TLS_H 1
#include "sm/generic.h"
#include "sm/error.h"
#include "sm/log.h"
#include "sm/str.h"
#include "sm/time.h"

#if MTA_USE_TLS
#include "sm/tlsstr.h"
#include "sm/tlscnf.h"
#include <openssl/ssl.h>
#include <openssl/err.h>
#include <openssl/bio.h>
#include <openssl/pem.h>
#include <openssl/x509.h>
#include <openssl/x509v3.h>
#ifndef HAVE_URANDOMDEV
# include <openssl/rand.h>
#endif

#define SM_X509_NAME_MAX	256
#define SM_TLS_NAME_MAX		1024
#define MTA_ALTNAMES_MAX	16
#define MTA_VERIFY_DEPTH_MAX	16

/* TLS library context */
struct tls_ctx_S
{
	sm_magic_T		 sm_magic;
	sm_log_ctx_P	 tls_lctx;
	sm_logconfig_P	 tls_lcfg;
	SSL_CTX			*tls_ssl_ctx;
	bool			 tls_clt_cache;
	tls_cnf_P		 tls_cnf;
	const EVP_MD	*tls_EVP_digest;
};

#if TLSL_LOG_DEFINES
sm_logcategory_T tlsl_lcats[] =
{
	{ "init",	0 },
	{ "config",	0 },
	{ "io",		0 },
	{ NULL,	0 }
};

sm_logmodule_T tlsl_lmods[] =
{
	{ "config",		0 },
	{ "io",			0 },
	{ NULL,		0 }
};

#else /* TLSL_LOG_DEFINES */

extern sm_logcategory_T tlsl_lcats[];
extern sm_logmodule_T tlsl_lmods[];

#endif /* TLSL_LOG_DEFINES */

#define TL_LCAT_INIT	(&tlsl_lcats[0])
#define TL_LCAT_COMM	(&tlsl_lcats[1])
#define TL_LCAT_IO		(&tlsl_lcats[2])

/* Backwards compatibility. */
#define TL_LCAT_GENERAL		TLGLOGCATEGORY_GENERAL

#define TL_LMOD_CONF	(&tlsl_lmods[0])
#define TL_LMOD_IO		(&tlsl_lmods[1])

void tlsl_log_init(sm_log_ctx_P _lctx);
void tlsl_log_setcontext(sm_log_ctx_P _lctx);

/* TLS information context */
struct tlsi_ctx_S
{
	/* pointer to library ctx, used in cb; currently only for logging */
	tls_ctx_P	 tlsi_tls;
	int			 tlsi_vrfy;
	int			 tlsi_cipher_bits;
	int			 tlsi_algs_bits;
	int			 tlsi_reused;
	sm_str_P	 tlsi_cipher;
	sm_str_P	 tlsi_version;
	sm_str_P	 tlsi_cert_subject;
	sm_str_P	 tlsi_cert_issuer;
	sm_str_P	 tlsi_cn_subject;
	sm_str_P	 tlsi_cn_issuer;
	sm_str_P	 tlsi_altnames[MTA_ALTNAMES_MAX];
	int			 tlsi_naltnames;
	sm_str_P	 tlsi_ca_subjects[MTA_VERIFY_DEPTH_MAX];
	sm_str_P	 tlsi_vrfy_err;
	sm_str_P	 tlsi_cert_fp;
	const char	*tlsi_log_field;	/* extra data to log: field */
	const char	*tlsi_log_value;	/* ...: value */
#if FFR_CERT_PINNING
	/* use tlscp_ctx_T? */
	uint		 tlsi_cert_mdlen;
	uchar		 tlsi_cert_md[EVP_MAX_MD_SIZE];
	int			 tlsi_hs_stat;
	uint32_t	 tlsi_hs_tricks;
#endif
#if FFR_CERT_PINNING || MTA_TLS_NOTAFTER
	time_T		 tlsi_notAfter_T;
#endif
#if MTA_TLS_NOTAFTER
	sm_str_P	 tlsi_notAfter;
#endif

};

/* what to do in the TLS initialization */
#define TLS_I_NONE		0x00000000	/* no requirements... */
/* need to make these "indexable" -- see smtls.c: TLS_S_CERT_OK(i) */
#define TLS_I_CERT(i)	(0x00000001 << (i << 1))	/* cert must usable */
#define TLS_I_KEY(i)	(0x00000002 << (i << 1))	/* key must usable */

#define TLS_IDX_RSA	0
#define TLS_IDX_DSA	1
#define TLS_IDX_EC	2

#define TLS_I_USE_KEY	TLS_I_KEY(TLS_IDX_RSA)	/* private key must usable */
#define TLS_I_USE_CERT	TLS_I_CERT(TLS_IDX_RSA)	/* certificate must be usable */
#define TLS_I_USE_DSA_KEY	TLS_I_KEY(TLS_IDX_DSA)	/* private DSA key must usable */
#define TLS_I_USE_DSA_CERT	TLS_I_CERT(TLS_IDX_DSA)	/* DSA certificate must be usable */
#define TLS_I_USE_EC_KEY	TLS_I_KEY(TLS_IDX_EC)	/* private EC key must usable */
#define TLS_I_USE_EC_CERT	TLS_I_CERT(TLS_IDX_EC)	/* EC certificate must be usable */

#define TLS_I_CERT_EX	0x00000040	/* cert must exist */
#define TLS_I_KEY_EX	0x00000080	/* key must exist */
#define TLS_I_CERTP_EX	0x00000100	/* CA cert path must exist */
#define TLS_I_CERTF_EX	0x00000200	/* CA cert file must exist */
#define TLS_I_LISTF_EX	0x00000400	/* CA list file must exist */
#define TLS_I_RSA_TMP	0x00000800	/* RSA TMP must be generated */

#define TLS_I_VRFY_PATH	0x00001000	/* load verify path must succeed */
#define TLS_I_VRFY_LOC	0x00002000	/* load verify default must succeed */
#define TLS_I_CACHE		0x00004000	/* require cache */
#define TLS_I_TRY_DH	0x00008000	/* try DH certificate */
#define TLS_I_REQ_DH	0x00010000	/* require DH certificate */
#define TLS_I_DHPAR_EX	0x00020000	/* require DH parameters */
#define TLS_I_DH512		0x00040000	/* generate 512bit DH param */
#define TLS_I_DH1024	0x00080000	/* generate 1024bit DH param */
#define TLS_I_DH2048	0x00100000	/* generate 2048bit DH param */
#define TLS_I_NO_VRFY	0x00200000	/* do not require authentication */
#define TLS_I_KEY_OUNR	0x00400000	/* Key must be other unreadable (NI) */
#define TLS_I_LDCLTCA	0x01000000	/* load CA list for clients */

/* require server cert */
#define TLS_I_SRV_CERT	 (TLS_I_CERT_EX | TLS_I_KEY_EX | \
			  TLS_I_KEY_OUNR | \
			  TLS_I_CERTP_EX | TLS_I_CERTF_EX | \
			  TLS_I_USE_KEY | TLS_I_USE_CERT)

/* server requirements */
#define TLS_I_SRV	(TLS_I_SRV_CERT | TLS_I_RSA_TMP | TLS_I_VRFY_PATH | \
			 TLS_I_VRFY_LOC | TLS_I_CACHE | TLS_I_TRY_DH | TLS_I_LDCLTCA)

/* client requirements */
#define TLS_I_CLT	(TLS_I_KEY_OUNR)

#define TLS_VRFY_OK	0	/* verification ok */
#define TLS_VRFY_NO	1	/* not verified */
#define TLS_VRFY_NOTR	2	/* verification not requested */
#define TLS_VRFY_FAIL	(-1)	/* verification failed */
#define TLS_VRFY_NONE	(-2)	/* no verification performed so far */

sm_ret_T sm_tls_init_library(tls_ctx_P *_ptls_ctx, bool _fipsmode);
sm_ret_T sm_tls_init(tls_ctx_P _tls_ctx, ulong _req, bool _srv, tls_cnf_P tls_cnf);
sm_ret_T sm_tlsversionprt(sm_file_T *_fp);
sm_ret_T sm_tlsversionok(void);
void	 tlslogerr(SSL *_ssl, sm_log_ctx_P _tls_lctx, char *_who);
void	 tls_set_verify(SSL_CTX *_ctx, SSL *_ssl, bool _vrfy);

sm_ret_T sm_ssl_loadcert(tls_ctx_P _tls_ctx, SSL *_ssl, bool _srv, const char *_keyfile, const char *_certfile);

/* flags for tls_get_info */
#define TLS_F_CERT_RQST	0x02	/* cert has been requested */

sm_ret_T tls_get_info(tls_ctx_P _tls_ctx, SSL *_ssl, uint _flags, tlsi_ctx_P _tlsi_ctx);
sm_ret_T sm_tls_get_cn(X509_NAME *_x509_name, int _nid, sm_str_P *_str);
sm_ret_T sm_tls_get_altname(X509 *_cert, sm_str_P _strs[], int _nstr);
int	 endtls(tls_ctx_P _tls_ctx, SSL *_ssl, char *_side);

sm_ret_T sm_set_tls_log(tls_ctx_P _tls_ctx, sm_file_T *_fp, int _fd, int _loglevel);

const char	*tls_vrfy2txt(int _vrfy);
sm_ret_T tls_set_md(tls_ctx_P _tls_ctx, tls_cnf_P _tls_cnf);

sm_ret_T tlsi_new(tlsi_ctx_P *_ptlsi_ctx, tls_ctx_P _tls_ctx);
sm_ret_T tlsi_free(tlsi_ctx_P _tlsi_ctx);
sm_ret_T tlsi_clr(tlsi_ctx_P _tlsi_ctx);


extern int TLSctxidx, TLSsslidx;

/* note: this uses a "fixed" error label: enomem */
#define TLS_GEN_PATHS(tls_cnf, confdir) do {	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_certfile,	\
		(tls_cnf).tlscnf_cert_abs, confdir,	\
		(tls_cnf).tlscnf_cert, SM_MTA_CERT, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_keyfile,	\
		(tls_cnf).tlscnf_key_abs, confdir,	\
		(tls_cnf).tlscnf_key, SM_MTA_KEY, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_dsa_certfile,	\
		(tls_cnf).tlscnf_dsa_cert_abs,	\
		confdir, (tls_cnf).tlscnf_dsa_cert, NULL, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_dsa_keyfile,	\
		(tls_cnf).tlscnf_dsa_key_abs,	\
		confdir, (tls_cnf).tlscnf_dsa_key, NULL, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_ec_certfile,	\
		(tls_cnf).tlscnf_ec_cert_abs,	\
		confdir, (tls_cnf).tlscnf_ec_cert, NULL, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_ec_keyfile,	\
		(tls_cnf).tlscnf_ec_key_abs,	\
		confdir, (tls_cnf).tlscnf_ec_key, NULL, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_cacertpath,	\
		(tls_cnf).tlscnf_certpath_abs,	\
		confdir, (tls_cnf).tlscnf_certpath, SM_MTA_CERTP, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_cacertfile,	\
		(tls_cnf).tlscnf_cacert_abs,	\
		confdir, (tls_cnf).tlscnf_cacert, SM_MTA_CACERT, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_calistfile,	\
		(tls_cnf).tlscnf_calist_abs,	\
		confdir, (tls_cnf).tlscnf_calist, SM_MTA_CACERT, enomem);	\
	} while (0)

/* copy of above without defaults */
#define TLS_GEN_PATHS_NODFLTS(tls_cnf, confdir) do {	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_certfile,	\
		(tls_cnf).tlscnf_cert_abs, confdir,	\
		(tls_cnf).tlscnf_cert, NULL, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_keyfile,	\
		(tls_cnf).tlscnf_key_abs, confdir,	\
		(tls_cnf).tlscnf_key, NULL, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_dsa_certfile,	\
		(tls_cnf).tlscnf_dsa_cert_abs,	\
		confdir, (tls_cnf).tlscnf_dsa_cert, NULL, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_dsa_keyfile,	\
		(tls_cnf).tlscnf_dsa_key_abs,	\
		confdir, (tls_cnf).tlscnf_dsa_key, NULL, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_ec_certfile,	\
		(tls_cnf).tlscnf_ec_cert_abs,	\
		confdir, (tls_cnf).tlscnf_ec_cert, NULL, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_ec_keyfile,	\
		(tls_cnf).tlscnf_ec_key_abs,	\
		confdir, (tls_cnf).tlscnf_ec_key, NULL, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_cacertpath,	\
		(tls_cnf).tlscnf_certpath_abs,	\
		confdir, (tls_cnf).tlscnf_certpath, NULL, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_cacertfile,	\
		(tls_cnf).tlscnf_cacert_abs,	\
		confdir, (tls_cnf).tlscnf_cacert, NULL, enomem);	\
	SM_GEN_MAP_PATH((tls_cnf).tlscnf_calistfile,	\
		(tls_cnf).tlscnf_calist_abs,	\
		confdir, (tls_cnf).tlscnf_calist, NULL, enomem);	\
	} while (0)

#define TLS_SET_XSTR(s, str) do {	\
		if ((s) != NULL) {					\
			int sl;	\
			sl = strlen(s);					\
			if ((str) == NULL)				\
				(str) = sm_str_new(NULL, sl + 16, SM_TLS_NAME_MAX);	\
			else						\
				sm_str_clr(str);			\
			if ((str) != NULL)				\
				(void) xtextify2((s), "<>\")", str);	\
		}							\
	} while (0)

#endif /* MTA_USE_TLS */

#endif /* SM_TLS_H */
