/*
 * Copyright (c) 2010 Sendmail, Inc. and its suppliers.
 *      All rights reserved.
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the LICENSE file which can be found at the top level of
 * the sendmail distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: msg.c,v 1.21 2011/03/04 16:23:05 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/memops.h"
#include "sm/time.h"
#include "sm/heap.h"
#include "sm/mta.h"
#include "sm/dkimdef.h"
#include "sm/dkim.h"
#include "dkim.h"

#if MTA_USE_DKIM

/*
**  DKIM_MSG_FREE -- free a DKIM msg context
**
**	Parameters:
**		dkim_msg -- DKIM msg context
**
**	Returns:
**		SM_SUCCESS
*/

sm_ret_T
dkim_msg_free(dkim_msg_P dkim_msg)
{
	if (dkim_msg == NULL)
		return SM_SUCCESS;
	SM_STR_FREE(dkim_msg->dkm_hdr_n);
	SM_STR_FREE(dkim_msg->dkm_hdr_v);
	SM_STR_FREE(dkim_msg->dkm_hdrs);
	SM_STR_FREE(dkim_msg->dkm_sighdr);
	SM_STR_FREE(dkim_msg->dkm_fromhdr);
	dkhl_free(&(dkim_msg->dkm_hhd));
	sm_free_size(dkim_msg, sizeof(*dkim_msg));
	return SM_SUCCESS;
}

/*
**  DKIM_MSG_NEW -- create a DKIM msg context
**
**	Parameters:
**		dkim_ctx -- DKIM context
**		pdkim_msg -- (pointer to) DKIM msg context (output)
**
**	Returns:
**		usual error code (ENOMEM)
*/

#define STR_NEW(str, l, m)	do { \
		str = sm_str_new(NULL, l, m);	\
		if (NULL == str) goto enomem;	\
	} while (0)

sm_ret_T
dkim_msg_new(dkim_ctx_P dkim_ctx, dkim_msg_P *pdkim_msg)
{
	sm_ret_T ret;
	dkim_msg_P dkim_msg;

	SM_IS_DKIM_CTX(dkim_ctx);
	SM_REQUIRE(pdkim_msg != NULL);

	dkim_msg = (dkim_msg_P) sm_zalloc(sizeof(*dkim_msg));
	if (dkim_msg == NULL)
		return sm_error_temp(SM_EM_DKIM, ENOMEM);

	/* sizes??? */
	STR_NEW(dkim_msg->dkm_hdr_n, 64, SM_MAXHDRLEN);
	STR_NEW(dkim_msg->dkm_hdr_v, DKIM_HDRLEN, dkim_ctx->dkc_cnf->dkimcnf_hdrvlen);

	/* - length of header name and standard additions... */
#if SM_MAXHDRLEN <= 256
# ERROR _SM_MAXHDRLEN too small
#endif
	STR_NEW(dkim_msg->dkm_hdrs, DKIM_HDRLEN, SM_MAXHDRLEN - 256);
	STR_NEW(dkim_msg->dkm_sighdr, DKIM_HDRLEN, SM_MAXHDRLEN);

	DKH_INIT(&(dkim_msg->dkm_hhd));

	dkim_msg->dkm_ctx = dkim_ctx;
	*pdkim_msg = dkim_msg;
	return SM_SUCCESS;

  enomem:
	ret = sm_error_temp(SM_EM_DKIM, ENOMEM);
	dkim_msg_free(dkim_msg);
	return ret;

}

/*
**  DKIM_MSG_OPEN -- set up DKIM msg context
**
**	Parameters:
**		dkim_ctx -- DKIM context
**		pdkim_msg -- (pointer to) DKIM msg context (output)
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
dkim_msg_open(dkim_ctx_P dkim_ctx, dkim_msg_P *pdkim_msg)
{
	sm_ret_T ret;
	dkim_msg_P dkim_msg;
	dkim_cnf_P dkim_cnf;

	SM_REQUIRE(dkim_ctx != NULL);
	SM_REQUIRE(pdkim_msg != NULL);
	dkim_msg = NULL;
	ret = dkim_msg_new(dkim_ctx, pdkim_msg);
	if (sm_is_err(ret))
		return ret;

	dkim_msg = *pdkim_msg;
	dkim_cnf = dkim_ctx->dkc_cnf;
	ret = dkim_open_write(&dkim_ctx, DKIM_SIGNM(dkim_cnf), &dkim_msg->dkm_hfp);
	if (sm_is_err(ret))
		goto error;

	dkim_msg->dkm_eoh_state = 0;
	dkim_msg->dkm_eot_state = SM_GOT_CRLF;
	dkim_msg->dkm_fhdr_state = SM_GOT_CRLF;

	return SM_SUCCESS;

  error:
	dkim_msg_free(dkim_msg);
	return ret;
}

/*
**  DKIM_MSG_CLOSE -- close DKIM msg context
**
**	Parameters:
**		dkim_msg -- DKIM msg context
**
**	Returns:
**		usual sm_error code
*/

sm_ret_T
dkim_msg_close(dkim_msg_P dkim_msg)
{
	if (dkim_msg->dkm_hfp != NULL) {
		dkim_close(dkim_msg->dkm_hfp, 0);
		dkim_msg->dkm_hfp = NULL;
	}
	if (dkim_msg->dkm_bfp != NULL) {
		dkim_close(dkim_msg->dkm_bfp, 0);
		dkim_msg->dkm_bfp = NULL;
	}
	dkim_msg_free(dkim_msg);
	return SM_SUCCESS;
}

#endif /* MTA_USE_DKIM */
