/*
 * Copyright (c) 2013 Claus Assmann
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the license/LICENSE.3C file which can be found at the
 * top level of this source code distribution.
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: tlsaltname.c,v 1.11 2014/01/27 16:45:37 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/ctype.h"
#include "sm/tls.h"

#if MTA_USE_TLS

/* Extract name from subjectAltName value */
static const char *
tls_get_name(const GENERAL_NAME * gn)
{
	int	len, i;
	const char *str;

	SM_REQUIRE(gn != NULL);
	SM_REQUIRE(gn->type == GEN_DNS);

	if (ASN1_STRING_type(gn->d.ia5) != V_ASN1_IA5STRING) {
	/* msg_warn("%s: %s: invalid ASN1 value type in subjectAltName",
		 myname, TLScontext->namaddr); */
		return NULL;
	}

	str = (const char *) ASN1_STRING_data(gn->d.ia5);
	len = ASN1_STRING_length(gn->d.ia5);

	/*
	 *  "remove" trailing NULs (except for one of course), those can happen
	 *  and are OK (not a sign of an attack)
	 */
	while (len > 0 && '\0' == str[len - 1])
			len--;

	/* just check for printable char for now */
	for (i = 0; i < len && ISPRINT(str[i]); i++)
			;
	if (str[i] != '\0' || len != i)
		return NULL;
	return str;
}

/*
**  SM_TLS_GET_ALTNAME -- extract AltName from an X509 name
**
**	Parameters:
**		x509_name -- X509 name
**		strs -- where to write result (output)
**		nstr -- size of str
**
**	Returns:
**		>0: number of entries created
**		error codes
**		sm_err_perm(SM_E_UNAVAIL): no subject_alt_name
**		sm_err_perm(SM_E_NOTFOUND): no matching entry
*/

sm_ret_T
sm_tls_get_altname(X509 *cert, sm_str_P strs[], int nstr)
{
	int r, i, n;
	const char *cs;
	const GENERAL_NAME *gn;
	STACK_OF(GENERAL_NAME) * gens;

	gens = X509_get_ext_d2i(cert, NID_subject_alt_name, 0, 0);
	if (NULL == gens)
		return sm_err_perm(SM_E_UNAVAIL);

	r = sk_GENERAL_NAME_num(gens);
	for (n = i = 0; i < r; ++i) {
		gn = sk_GENERAL_NAME_value(gens, i);
		if (gn == NULL) /* oops? */
			 continue;
		if (gn->type != GEN_DNS)
			 continue;

		cs = tls_get_name(gn);
#if MTA_TLS_DEBUG > 10
		sm_io_fprintf(smioerr, "sev=DBG, func=%s, i=%d, cs=%s\n", SMFCT, i, cs);
#endif
		if (cs != NULL && n < nstr) {
			TLS_SET_XSTR(cs, strs[n]);
			++n;
		}
	}

	sk_GENERAL_NAME_pop_free(gens, GENERAL_NAME_free);

	return n;
}
#endif /* MTA_USE_TLS */
