/*
 *  Code based on tls_dane.c from postfix which has the following
 * LICENSE
 *	This software is free. You can do with it whatever you want.
 *	The original author kindly requests that you acknowledge
 *	the use of his software.
 * AUTHOR(S)
 *	Wietse Venema
 *	IBM T.J. Watson Research
 *	P.O. Box 704
 *	Yorktown Heights, NY 10598, USA
 *
 *	Viktor Dukhovni
 */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: tlsdane.c,v 1.7 2013/12/28 21:04:46 ca Exp $")
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/ctype.h"
#include "sm/dns.h"
#include "sm/tls.h"
#include "sm/tlsdane.h"
#include "sm/str.h"

#if MTA_USE_TLS && MTA_TLS_DANE

/*
 * https://www.iana.org/assignments/dane-parameters/dane-parameters.xhtml
 */

struct digest_info_S
{
	const char		*di_alg;		/* OpenSSL name */
	const EVP_MD	*di_md;			/* OpenSSL EVP handle */
	int				 di_len;		/* digest octet length */
	int				 di_pref;		/* tls_dane_digests index or -1 */
	uint8_t			 di_dane_id;	/* IANA id */
};
typedef struct digest_info_S digest_info_T, *digest_info_P;

#define MAXDIGESTS 256			/* RFC limit */
digest_info_T digest_table[] = {
	{"full",	NULL,	 0,	 0,	DNS_TLSA_MATCHING_TYPE_NO_HASH_USED},
	{"sha256",	NULL,	-1,	-1,	DNS_TLSA_MATCHING_TYPE_SHA256},
	{"sha512",	NULL,	-1,	-1, DNS_TLSA_MATCHING_TYPE_SHA512},
	{NULL,		NULL,	 0,	 0,	0}
};


/* dane_digest_info - locate digest_table entry for given IANA id */
static digest_info_P
dane_digest_info(uint8_t dane_id)
{
	digest_info_P di;

	for (di = digest_table; di->di_alg != NULL; ++di)
		if (di->di_dane_id == dane_id)
			return di;
	return NULL;
}

/* dane_init - initialize DANE parameters */
sm_ret_T
dane_init(dane_ctx_P *dane_ctx)
{
	digest_info_P di;

	for (di = digest_table; di->di_alg != NULL; ++di) {
		if (di->di_len != -1) continue;	/* don't initialize this */
		di->di_md = EVP_get_digestbyname(di->di_alg);
		if (NULL == di->di_md) {
			sm_io_fprintf(smioerr, "dane_init: can't find alg=%s\n", di->di_alg);
			continue;
		}
		di->di_len = EVP_MD_size(di->di_md);
		if (di->di_len <= 0 || di->di_len > EVP_MAX_MD_SIZE) {
			sm_io_fprintf(smioerr, "dane_init: bogus di_len=%d for alg=%s\n", di->di_len, di->di_alg);
			continue;
		}
	}
	ERR_clear_error();

	return SM_SUCCESS;
}

#endif /* MTA_USE_TLS && MTA_TLS_DANE */
