/*
 * Copyright (c) 2010 Claus Assmann
 *
 * By using this file, you agree to the terms and conditions set
 * forth in the license/LICENSE.3C file which can be found at the
 * top level of this source code distribution.
 */

/* test IQDB error handling */

#include "sm/generic.h"
SM_RCSID("@(#)$Id: t-iqdb-0.c,v 1.17 2014/01/30 15:43:32 ca Exp $")
#include "sm/common.h"
#include "sm/assert.h"
#include "sm/error.h"
#include "sm/sysexits.h"
#include "sm/version.h"
#include "sm/io.h"
#include "sm/qmgr.h"
#include "sm/qmgr-int.h"
#include "qmgr.h"
#include "sm/reccom.h"
#include "log.h"
#include "sm/test.h"

/* fixme: where to define? */
cdb_id_P null_cdb_id;

#if QMGR_TEST

static void
iqdb_print(const char *key, const void *value, void *ctx, void *walk_ctx, uint type)
{
	qss_ta_P qss_ta;
	qss_sess_P qss_sess;
	qss_rcpt_P qss_rcpt;
	qmgr_ctx_P qmgr_ctx;

	qmgr_ctx = (qmgr_ctx_P) ctx;
	SM_IS_QMGR_CTX(qmgr_ctx);

	switch (type) {
	  case IQDB_T_SESS:
		qss_sess = (qss_sess_P) value;
		SM_IS_QS_SE(qss_sess);
		sm_log_write(qmgr_ctx->qmgr_lctx,
			QM_LCAT_CLEANUP, QM_LMOD_CLEANUP,
			SM_LOG_DEBUG, 7,
			"sev=DEBUG, func=iqdb_print, qss_sess=%p, sess_id=%s"
			, qss_sess, qss_sess->qsses_id);

		break;

	  case IQDB_T_TA:
		qss_ta = (qss_ta_P) value;
		SM_IS_QS_TA(qss_ta);
		sm_log_write(qmgr_ctx->qmgr_lctx,
			QM_LCAT_CLEANUP, QM_LMOD_CLEANUP,
			SM_LOG_DEBUG, 7,
			"sev=DEBUG, func=iqdb_print, qss_ta=%p, ta_id=%s"
			, qss_ta, qss_ta->qssta_id);
		break;

	  case IQDB_T_RCPT:
		qss_rcpt = (qss_rcpt_P) value;
		SM_IS_QS_RCPT(qss_rcpt);
		sm_log_write(qmgr_ctx->qmgr_lctx,
			QM_LCAT_CLEANUP, QM_LMOD_CLEANUP,
			SM_LOG_DEBUG, 7,
			"sev=DEBUG, func=iqdb_print, qss_rcpt=%p, rcpt_id=%s"
			, qss_rcpt, qss_rcpt->qsr_id);
		break;
	}
}

/*
**  T_QM_FR_SS_REACT -- Decode data received from SMTPS
**
**	Parameters:
**		qss_ctx -- QMGR/SMTPS context
**
**	Returns:
**		usual sm_error code
**
**	Locking:
**		XXX who locks qss_ctx?
**
**	Called by: qm_fr_ss
*/

static sm_ret_T
t_qm_fr_ss_react(qss_ctx_P qss_ctx, uint iqdbsize, uint nrcpts)
{
	uint32_t v, l, rt, tl;
	off_t off;
	sm_ret_T ret, ret2, rv;
#if FFR_AVLSLOTS
# define NO_AVAIL_SLOTS , 0
#else
# define NO_AVAIL_SLOTS
#endif
	qss_status_T oldstatus;
	bool inwaitq;
	int r;
	uint32_t fct_state;
	sm_rcb_P rcb;
	sm_rcbe_P rcbe;
	qmgr_ctx_P qmgr_ctx;
	sm_evthr_task_P tsk;

#define FST_QSS_TA_LOCKED	0x0001	/* qss_ta is locked */

	SM_IS_QSS_CTX(qss_ctx);
	ret = rv = SM_SUCCESS;
	inwaitq = false;
	fct_state = 0;
	rcbe = NULL;
	tsk = NULL;

	/* Decode rcb */
	rcb = qss_ctx->qss_com.rcbcom_rdrcb;
	ret = sm_rcb_open_dec(rcb);
	SM_TEST(SM_SUCCESS == ret);
	if (sm_is_err(ret)) {
		rv = ret;
		goto error;
	}
	qmgr_ctx = qss_ctx->qss_qmgr_ctx;
	SM_IS_QMGR_CTX(qmgr_ctx);

	/* Total length of record */
	ret = sm_rcb_getuint32(rcb, &tl);
	SM_TEST(SM_SUCCESS == ret);
	if (sm_is_err(ret) || tl > QM_SS_REC_LEN_MAX || tl > sm_rcb_getlen(rcb)) {
		rv = sm_is_err(ret) ? ret : sm_error_perm(SM_EM_Q_SS2Q, SM_E_RCB2LONG);
		goto err2;
	}

	/* Decode data, act accordingly... */

	/* Protocol header: version */
	ret = sm_rcb_get3uint32(rcb, &l, &rt, &v);
	SM_TEST(SM_SUCCESS == ret);
	if (sm_is_err(ret)) {
		rv = ret;
		goto err2;
	}
	if (l != 4 || rt != RT_PROT_VER || v != PROT_VER_RT) {
		rv = sm_error_perm(SM_EM_Q_SS2Q, SM_E_PR_V_MISM);
		goto err2;
	}

	/* SMTPS (new/existing/close) ID */
	ret = sm_rcb_get3uint32(rcb, &l, &rt, &v);
	SM_TEST(SM_SUCCESS == ret);
	if (sm_is_err(ret) || l != 4) {
		rv = sm_is_err(ret) ? ret : sm_error_perm(SM_EM_Q_SS2Q, SM_E_PR_ERR);
		goto err2;
	}

	/* rt == RT_S2Q_ID is the only case in which we continue here */
	oldstatus = qss_ctx->qss_status;

	/* what's next? */
	ret = sm_rcb_get2uint32(rcb, &l, &rt);
	SM_TEST(SM_SUCCESS == ret);
	if (sm_is_err(ret)) {
		rv = ret;
		goto err2;
	}
	if (RT_S2Q_1TAID == rt) {
		qss_ta_P qss_ta;
		uint rcptcnt;
		bool aborted;

		aborted = false;
		if (l != SMTP_STID_SIZE) {
			rv = sm_error_perm(SM_EM_Q_SS2Q, SM_E_PR_ERR);
			goto err2;
		}

		/* allocate new transaction? rpool? */
		ret = qss_ta_new(&qss_ta, qss_ctx, NULL);
		SM_TEST(SM_SUCCESS == ret);
		if (sm_is_err(ret))
			goto err2;
		qss_ta->qssta_st_time = evthr_time(qmgr_ctx->qmgr_ev_ctx);

		/* transaction id */
		ret = sm_rcb_getn(rcb, (uchar *) qss_ta->qssta_id, l);
		SM_TEST(SM_SUCCESS == ret);
		if (sm_is_err(ret)) {
			rv = ret;
			goto errntaid;
		}
		QM_LEV_DPRINTFC(QDC_S2Q, 3, (QM_DEBFP, "sev=DBG, func=qm_fr_ss_react, id=%d, 1ta-id=%s\n", qss_ctx->qss_id, qss_ta->qssta_id));
		ret = sm_getsmtpsid(qss_ta->qssta_id);

		if (sm_idcmp(qss_ctx->qss_sessta_id, qss_ta->qssta_id) < 0)
			SESSTA_COPY(qss_ctx->qss_sessta_id, qss_ta->qssta_id);

		ret = sm_rcb_get2uint32(rcb, &l, &rt);
		SM_TEST(SM_SUCCESS == ret);
		if (sm_is_err(ret) || rt != RT_S2Q_MAIL || l >= tl) {
			rv = sm_is_err(ret) ? ret : sm_error_perm(SM_EM_Q_SS2Q, SM_E_PR_ERR);
			goto errntaid;
		}
		ret = qss_mail_new(qss_ta);
		SM_TEST(SM_SUCCESS == ret);
		if (sm_is_err(ret))
			goto errntaid;
		ret = sm_rcb_getnstr(rcb, &qss_ta->qssta_mail->qsm_pa, l);
		SM_TEST(SM_SUCCESS == ret);
		if (sm_is_err(ret))
			goto errntaid;
		QM_LEV_DPRINTFC(QDC_S2Q, 2, (QM_DEBFP, "sev=DBG, func=qm_fr_ss_react, id=%d, 1ta-id=%s, mail from=%S, l=%d\n", qss_ctx->qss_id, qss_ta->qssta_id, qss_ta->qssta_mail->qsm_pa, l));

		ret = sm_rcb_peek2uint32(rcb, &l, &rt);
		SM_TEST(SM_SUCCESS == ret);
		if (sm_is_success(ret) && 4 == l && RT_S2Q_MAIL_FL == rt) {
			ret = sm_rcb_get3uint32(rcb, &l, &rt, &v);
			SM_TEST(SM_SUCCESS == ret);
			if (sm_is_err(ret) || l != 4 || rt != RT_S2Q_MAIL_FL) {
				rv = sm_is_err(ret) ? ret
					: sm_error_perm(SM_EM_Q_SS2Q, SM_E_PR_ERR);
				goto errntaid;
			}

			/* "translate" flags */
			if (SM_IS_FLAG(v, SM_TA_FL_VERP))
				QSS_TA_SET_FLAG(qss_ta, QSS_TA_FL_VERP);
		}

		ret = sm_rcb_get2uint32(rcb, &l, &rt);
		SM_TEST(SM_SUCCESS == ret);
		rcptcnt = 0;
		ret2 = SMTP_R_OK;
		while (RT_S2Q_RCPT_IDX == rt) {
			rcpt_idx_T rcpt_idx;
			qss_rcpt_P qss_rcpt;
			sm_str_P rcpt_pa;

			qss_rcpt = NULL;
			if (l != 4) {
				rv = sm_error_perm(SM_EM_Q_SS2Q, SM_E_PR_ERR);
				goto errntaid;
			}
			ret = sm_rcb_getuint32(rcb, &v);
			SM_TEST(SM_SUCCESS == ret);
			if (sm_is_err(ret)) {
				rv = ret;
				goto errntaid;
			}
			rcpt_idx = v;
			ret = sm_rcb_get2uint32(rcb, &l, &rt);
			SM_TEST(SM_SUCCESS == ret);
			if (sm_is_err(ret) || rt != RT_S2Q_RCPT || l > tl) {
				rv = sm_is_err(ret) ? ret
					: sm_error_perm(SM_EM_Q_SS2Q, SM_E_PR_ERR);
				goto errntaid;
			}

			ret = sm_rcb_getnstr(rcb, &rcpt_pa, l);
			SM_TEST(SM_SUCCESS == ret);
			if (sm_is_err(ret))
				goto errntaid;

			/* add recipient to list */
			ret = qss_rcpts_new(qss_ta, &rcpt_pa, rcpt_idx, &qss_rcpt);
			SM_TEST(SM_SUCCESS == ret);
			if (sm_is_err(ret))
				goto errntaid;
			QSRCPT_SET_FLAG(qss_rcpt, QSRCPT_FL_TA);
			QM_LEV_DPRINTFC(QDC_S2Q, 2, (QM_DEBFP, "sev=DBG, func=qm_fr_ss_react, id=%d, qss_ta-id=%s, rcpt_idx=%d, rcptaddr=%S, l=%d, nrcpts=%d\n",
qss_ctx->qss_id, qss_ta->qssta_id, rcpt_idx, qss_rcpt->qsr_pa, l, qss_ta->qssta_rcpts_tot));

			/* add recipient to iqdb, ibdb */
			/* XXX split here: this function will be async. */
			ret2 = qm_ss_rcptid(qss_ctx, qss_ta, tsk, qss_rcpt);
			qss_rcpt = NULL; /* qm_ss_rcptid took care of it */

			if (iqdbsize == rcptcnt)
				SM_TEST(SMTP_R_TEMP == ret2);

			/* qss_rcpt will be in qss_ta only if accepted */
			if (sm_is_err(ret2)) {
				QM_LEV_DPRINTFC(QDC_S2Q, 0, (QM_DEBFP, "sev=ERROR, func=qm_fr_ss_react, qm_ss_rcptid=%r\n", ret2));
				ret = ret2;
				goto errntaid;
			}
			if (ret2 != SMTP_R_OK) {
				QM_LEV_DPRINTFC(QDC_S2Q, 0, (QM_DEBFP, "sev=ERROR, func=qm_fr_ss_react, qm_ss_rcptid=%r\n", ret2));
				aborted = true;
				goto ntaidreply;
			}

			ret = sm_rcb_get2uint32(rcb, &l, &rt);
			SM_TEST(SM_SUCCESS == ret);
			rcptcnt++;
		}

		if (rt != RT_S2Q_CDBID || l > tl) {
			QM_LEV_DPRINTFC(QDC_S2Q, 0, (QM_DEBFP, "sev=ERROR, func=qm_fr_ss_react, rt=%#x, l=%u, tl=%u\n", rt, l, tl));
			rv = sm_error_perm(SM_EM_Q_SS2Q, SM_E_PR_ERR);
			goto errntaid;
		}

		ret = sm_rcb_getncstr(rcb, &qss_ta->qssta_cdb_id, l);
		SM_TEST(SM_SUCCESS == ret);
QM_LEV_DPRINTFC(QDC_S2Q, 4, (QM_DEBFP, "sev=DBG, func=qm_fr_ss_react, cdb_id=%C, ret=%r\n", qss_ta->qssta_cdb_id, ret));
		if (sm_is_err(ret)) {
			/*
			**  XXX Throttle SMTPS? Send error back saying that the
			**  transaction can't be accepted because QMGR is out of memory?
			*/

			if (qss_ta != NULL) {
				/*
				 *  Remove from DBs? It's not in IBDB nor in AQ.
				 *  Use qm_ss_dtaid() instead?
				 */

				ret = qss_ta_free(qss_ta, false, QSS_TA_FREE_ALWAYS, QSS_RMFIQDB);
				SM_TEST(SM_SUCCESS == ret);
				qss_ta = NULL;
			}
			goto errntaid;
		}
		ret = sm_rcb_get3off_t(rcb, &l, &rt, &off);
		SM_TEST(SM_SUCCESS == ret);
		if (sm_is_err(ret) || rt != RT_S2Q_SIZE_B || l != SIZEOF_OFF_T) {
			rv = sm_is_err(ret) ? ret
				: sm_error_perm(SM_EM_Q_SS2Q, SM_E_PR_ERR);
			goto errntaid;
		}
		qss_ta->qssta_msg_sz_b = off;
		QM_LEV_DPRINTFC(QDC_S2Q, 3, (QM_DEBFP, "sev=DBG, func=qm_fr_ss_react, id=%d, size=%lu\n", qss_ctx->qss_id, (ulong) off));

  ntaidreply:
		ret = qm_rcbcom_prerep(qmgr_ctx, &qss_ctx->qss_com, tsk, aborted, &rcbe);
		SM_TEST(SM_SUCCESS == ret);
		if (sm_is_err(ret)) {
			QM_LEV_DPRINTFC(QDC_S2Q, 0, (QM_DEBFP, "sev=ERROR, func=qm_fr_ss_react, qss_ta=%p, qm_rcbcom_prerep=%r\n", qss_ta, ret));
			goto errntaid;
		}
		inwaitq = true;

		r = pthread_mutex_lock(&qss_ta->qssta_mutex);
		SM_LOCK_OK(r);
		if (r != 0) {
			QM_LEV_DPRINTFC(QDC_S2Q, 0, (QM_DEBFP, "sev=ERROR, func=qm_fr_ss_react, qss_ta=%p, lock=%d\n", qss_ta, r));
			goto errntaid;
		}
		SM_SET_FLAG(fct_state, FST_QSS_TA_LOCKED);

		/* all data read, store it internally, prepare reply */

		if (SMTP_R_OK == ret2) {
			ret2 = qm_ss_1taid(qss_ctx, qss_ta);
			QM_LEV_DPRINTFC(QDC_S2Q, 1, (QM_DEBFP, "sev=DBG, func=qm_fr_ss_react, id=%d, 1ta-id=%s, qm_ss_1taid=%r\n", qss_ctx->qss_id, qss_ta->qssta_id, ret2));
			SM_TEST(!sm_is_err(ret2));
			if (sm_is_err(ret2))
				goto errntaid;
		}

		if (iqdbsize == nrcpts + 1)
			SM_TEST(SMTP_R_SSD == ret2);

		if (SMTP_R_OK == ret2) {
			ret2 = qm_ss_ctaid(qss_ctx, qss_ta);
			if (sm_is_err(ret2)) {
				QM_LEV_DPRINTFC(QDC_S2Q, 1, (QM_DEBFP, "sev=ERROR, func=qm_fr_ss_react, qm_ss_ctaid=%r, id=%d, cta-id=%s, cdb=%C\n", ret2, qss_ctx->qss_id, qss_ta->qssta_id, qss_ta->qssta_cdb_id));
				goto errcdb2;
			}
		}

		/*
		**  If we get an return code other than SMTP_R_OK
		**  then we can immediately tell SMTPS about;
		**  we don't need to schedule this for a group commit.
		*/

		if (ret2 != SMTP_R_OK) {
			sm_log_write(qss_ctx->qss_qmgr_ctx->qmgr_lctx,
				QM_LCAT_SMTPS, QM_LMOD_FROM_SMTPS,
				SM_LOG_INFO, 8,
				"sev=INFO, func=qm_fr_ss_react, smtps_id=%d, ss_ta=%s, stat=%r",
				qss_ctx->qss_id, qss_ta->qssta_id, ret2);

			ret = qm_2ss_ctaid(qss_ctx, qss_ta, rcbe, ret2 NO_AVAIL_SLOTS);
			SM_TEST(SM_SUCCESS == ret);
			if (sm_is_err(ret))
				goto errcdb2;
			ret = sm_rcbcom_endrep(&qss_ctx->qss_com, tsk, false, &rcbe);
			SM_TEST(SM_SUCCESS == ret);
			/* ret checked below */
			QM_LEV_DPRINTFC(QDC_S2Q, 6, (QM_DEBFP, "sev=DBG, func=qm_fr_ss_react, 3 evthr_en_wr=%p, done=%r\n", &qss_ctx->qss_com, ret));

			/* need to go to error handling section to remove all TA data */
		}
		else {
#if QMGR_STATS
			/* not protected by mutex */
			qmgr_ctx->qmgr_rcpts_rcvd += qss_ta->qssta_rcpts_tot;
			++qmgr_ctx->qmgr_tas_rcvd;
#endif

			/* don't need rcbe here: a reply scheduled in qm_ss_schedctaid() */
			sm_rcbe_free(rcbe);
			rcbe = NULL;
		}
		if (sm_is_err(ret))
			goto errcdb2;
		SM_ASSERT(SM_IS_FLAG(fct_state, FST_QSS_TA_LOCKED));
		r = pthread_mutex_unlock(&qss_ta->qssta_mutex);
		if (r != 0) {
			QM_LEV_DPRINTFC(QDC_S2Q, 0, (QM_DEBFP, "sev=ERROR, func=qm_fr_ss_react, qss_ta=%p, unlock=%d\n", qss_ta, r));
			SM_ASSERT(0 == r);
		}
		SM_CLR_FLAG(fct_state, FST_QSS_TA_LOCKED);

		if (ret2 != SMTP_R_OK && qss_ta != NULL) {
			ret = qm_ss_dtaid(qss_ctx, qss_ta, QSS_TA_FREE_ALWAYS, QSS_RMFIQDB);
			if (sm_is_err(ret))
				QM_LEV_DPRINTFC(QDC_S2Q, 0, (QM_DEBFP, "sev=ERROR, func=qm_fr_ss_react, qss_ss_dtaid=%r\n", ret));
			qss_ta = NULL;
		}

		return QMGR_R_ASYNC;

	/* check: proper cleanup functionality??? */
  errcdb2:
		SM_ASSERT(SM_IS_FLAG(fct_state, FST_QSS_TA_LOCKED));
		r = pthread_mutex_unlock(&qss_ta->qssta_mutex);
		if (r != 0) {
			QM_LEV_DPRINTFC(QDC_S2Q, 0, (QM_DEBFP, "sev=ERROR, func=qm_fr_ss_react, qss_ta=%p, unlock=%d\n", qss_ta, r));
			SM_ASSERT(0 == r);
		}
		SM_CLR_FLAG(fct_state, FST_QSS_TA_LOCKED);

		QM_LEV_DPRINTFC(QDC_S2Q, 0, (QM_DEBFP, "sev=ERROR, func=qm_fr_ss_react, id=%d, cta-id=%s, ret=%r\n", qss_ctx->qss_id, qss_ta->qssta_id, ret));
		sm_rcbe_free(rcbe);
		rcbe = NULL;
		if (qss_ta != NULL) {
			/*
			**  XXX Remove from iqdb? XXX Really? Always?
			**  Maybe use the flags to denote what should happen with qss_ta?
			**  (indicate the "progress", i.e., where is qss_ta stored and
			**  whether something bad happened such that it must be freed).
			*/

/*
			(void) qss_ta_free(qss_ta, false, QSS_TA_FREE_ALWAYS, QSS_RMFIQDB);
*/
			(void) qm_ss_dtaid(qss_ctx, qss_ta, QSS_TA_FREE_ALWAYS, QSS_RMFIQDB);
			qss_ta = NULL;
		}
		sm_log_write(qss_ctx->qss_qmgr_ctx->qmgr_lctx,
			QM_LCAT_SMTPS, QM_LMOD_FROM_SMTPS,
			SM_LOG_ERR, 4,
			"sev=ERROR, func=qm_fr_ss_react, ss_ta=%s, close_ss_ta=%m",
			qss_ta != NULL ? qss_ta->qssta_id : "unknown", ret);
		goto err2;

	/* check: proper cleanup functionality??? */
  errntaid:
		if (qss_ta != NULL && QSS_TA_IS_FLAG(qss_ta, QSS_TA_FL_IQDB)) {
			(void) iqdb_trans_rm(qmgr_ctx->qmgr_iqdb, qss_ta->qssta_id,
					SMTP_STID_SIZE, qss_ta, THR_LOCK_UNLOCK);
			QM_LEV_DPRINTFC(QDC_IQDB, 5, (QM_DEBFP, "sev=DBG, func=qm_fr_ss_react, ta_id=%s, iqdb_trans_rm=?\n", qss_ta->qssta_id));
			QSS_TA_SET_FLAG(qss_ta, QSS_TA_FL_IQDB_RM);
		}
		if (SM_IS_FLAG(fct_state, FST_QSS_TA_LOCKED)) {
			r = pthread_mutex_unlock(&qss_ta->qssta_mutex);
			if (r != 0) {
				QM_LEV_DPRINTFC(QDC_S2Q, 0, (QM_DEBFP, "sev=ERROR, func=qm_fr_ss_react, qss_ta=%p, unlock=%d\n", qss_ta, r));
				SM_ASSERT(0 == r);
			}
			SM_CLR_FLAG(fct_state, FST_QSS_TA_LOCKED);
		}
		if (qss_ta != NULL) {
			/* qss_ta is not in any DB */
			if (!QSS_TA_OK_FREE(qss_ta->qssta_flags))
				QM_LEV_DPRINTFC(QDC_S2Q, 0, (QM_DEBFP, "sev=DBG, func=qm_fr_ss_react, ERROR=unexpected_state, flags=%#x, id=%d, qss_ta-id=%s\n", qss_ta->qssta_flags, qss_ctx->qss_id, qss_ta->qssta_id));

			(void) qm_ss_dtaid(qss_ctx, qss_ta, QSS_TA_FREE_ALWAYS, QSS_RMFIQDB);
			qss_ta = NULL;
		}
		sm_log_write(qss_ctx->qss_qmgr_ctx->qmgr_lctx,
			QM_LCAT_SMTPS, QM_LMOD_FROM_SMTPS,
			SM_LOG_ERR, 4,
			"sev=ERROR, func=qm_fr_ss_react, 1ta_new=%r", ret);
		goto err2;
	}
	else
		/* error */;

	if (!inwaitq) {
		ret = sm_rcb_close_dec(qss_ctx->qss_com.rcbcom_rdrcb);
		(void) sm_rcb_open_rcv(qss_ctx->qss_com.rcbcom_rdrcb);
	}
	if (ret == SM_SUCCESS && inwaitq)
		return QMGR_R_ASYNC;
	return rv;

	/* preserve original error code! */
  err2:
	/* use rcb functions that don't do check the state */
	if (!inwaitq)
		(void) sm_rcb_close_decn(qss_ctx->qss_com.rcbcom_rdrcb);
  error:
	/* open rcb for receiving next record */
	if (!inwaitq)
		(void) sm_rcb_open_rcvn(qss_ctx->qss_com.rcbcom_rdrcb);

	QM_LEV_DPRINTFC(QDC_S2Q, 0, (QM_DEBFP, "sev=DBG, func=qm_fr_ss_react, id=%d, stat=%r, error out, rt=%#x, v=%d, ret=%r, rv=%r, inwaitq=%d\n",
qss_ctx->qss_id, qss_ctx->qss_status, rt, v, ret, rv, inwaitq));
	if (rcbe != NULL)
		sm_rcbe_free(rcbe);
	if (inwaitq) {
		if (sm_is_err(rv))	/* shouldn't happen! */
			QM_LEV_DPRINTFC(QDC_S2Q, 0, (QM_DEBFP, "sev=ERROR, func=qm_fr_ss_react, rv=%r, inwaitq=%d\n", rv, inwaitq));
		return QMGR_R_ASYNC;
	}
	return rv;
#undef FST_QSS_TA_LOCKED
}

sm_ret_T
t_sm_s2q_1taid(sm_rcb_P rcb, uint nrcpts)
{
	sm_ret_T ret;
	uint ui;
	sm_str_P ssse_wr;
	sessta_id_T ssta_id;

	ret = -1;
	ssse_wr = sm_str_new(NULL, 256, 1024);
	SM_TEST_E(ssse_wr != NULL);
	sm_str_clr(ssse_wr);
	ret = sm_str_scat(ssse_wr, "<sender@example.com>");
	sm_snprintf(ssta_id, SMTP_STID_SIZE, SMTPS_STID_FORMAT, 1, 2);

	ret = sm_rcb_putrec(rcb, RCB_PUTR_FIRST|RCB_PUTR_OPEN, 0, -1,
		SM_RCBV_INT, RT_PROT_VER, PROT_VER_RT,
		SM_RCBV_INT, RT_S2Q_ID, 1,
		SM_RCBV_BUF, RT_S2Q_1TAID, ssta_id, SMTP_STID_SIZE,
		SM_RCBV_STR, RT_S2Q_MAIL, ssse_wr,
		SM_RCBV_END);
	SM_TEST_E(SM_SUCCESS == ret);

	/*
	 *  TODO: What's the number of rcpts? ssta_rcpts_ok?
	 */

	for (ui = 0; ui < nrcpts; ui++) {
		sm_str_clr(ssse_wr);
		ret = sm_strprintf(ssse_wr, "<rcpt%u@example.net>", ui);
		ret = sm_rcb_putv(rcb, 0,
			SM_RCBV_INT, RT_S2Q_RCPT_IDX, ui,
			SM_RCBV_STR, RT_S2Q_RCPT, ssse_wr,
			SM_RCBV_END);
		SM_TEST_E(SM_SUCCESS == ret);
		sm_str_clr(ssse_wr);
	}
	SM_TEST_E(SM_SUCCESS == ret);

	ret = sm_rcb_putv(rcb, 0,
		SM_RCBV_BUF, RT_S2Q_CDBID, ssta_id, SMTP_STID_SIZE,
		SM_RCBV_OFF, RT_S2Q_SIZE_B, (off_t) 1234,
		SM_RCBV_END);
	SM_TEST_E(SM_SUCCESS == ret);

	ret = sm_rcb_close_enc(rcb);
	SM_TEST_E(SM_SUCCESS == ret);

  error:
	return ret;
}

/*
**  T_IQDB -- QMGR loop; starts event thread loop
**
**	Parameters:
**		qmgr_ctx -- QMGR context
**
**	Returns:
**		usual sm_error code
*/

static sm_ret_T
t_iqdb(qmgr_ctx_P qmgr_ctx, uint iqdbsize, uint nrcpts)
{
	sm_ret_T ret;
	qss_ctx_P qss_ctx;

	SM_IS_QMGR_CTX(qmgr_ctx);

	ret = -1;
	SM_TEST_E(QMGR_ST_START == qmgr_ctx->qmgr_status);
	qmgr_ctx->qmgr_status = QMGR_ST_OK;

	ret = qss_ctx_new(qmgr_ctx, &qss_ctx);
	SM_TEST_E(SM_SUCCESS == ret);

	/* set up rcb: fill it with test data */
	ret = t_sm_s2q_1taid(qss_ctx->qss_com.rcbcom_rdrcb, nrcpts);
	SM_TEST_E(SM_SUCCESS == ret);

	/* start tests here... */
	ret = t_qm_fr_ss_react(qss_ctx, iqdbsize, nrcpts);
	SM_TEST_E(!sm_is_err(ret));
	return SM_SUCCESS;

  error:
	return ret;
}

static void
usage(const char *prg)
{
	sm_io_fprintf(smioerr, "usage: %s [options]\n", prg);
	exit(EX_USAGE);
}

/*
**  MAIN -- QMGR
**
**	Parameters:
**		argc -- number of arguments
**		argv -- vector of arguments
**
**	Returns:
**		exit code
*/

int
main(int argc, char *argv[])
{
	sm_ret_T ret;
	int c, exitcode;
	qmgr_ctx_T qmgr_ctx;
	char *prg;
	uint iqdbsize, nrcpts;
#if QMGR_DEBUG
	uint u;
#endif

	nrcpts =33;
	iqdbsize = 32;
	prg = argv[0];
	exitcode = 0;
	if (getuid() == 0 || geteuid() == 0) {
		sm_io_fprintf(smioerr, SM_DONTRUNASROOT "\n", prg);
		exit(EX_USAGE);
	}

	while ((c = getopt(argc, argv, "i:r:V")) != -1) {
		switch (c) {
		  case 'i':
			iqdbsize = (uint) strtoul(optarg, NULL, 0);
			break;
		  case 'r':
			nrcpts = (uint) strtoul(optarg, NULL, 0);
			break;
		  case 'V':
/*
			++Verbose;
*/
			break;
		  default:
			usage(prg);
		}
	}

	sm_test_begin(argc, argv, "test iqdb overflow 0");
	/* basic initialization */
	ret = qmgr_init0(&qmgr_ctx);
	SM_TEST(SM_SUCCESS == ret);
	if (sm_is_err(ret))
		goto error0;
#if SM_HEAP_CHECK
	SmHeapCheck = 1;
	if (HEAP_CHECK)
		sm_heap_report(smioerr, 3);
#endif

	/* read config */
	ret = qmgr_rdcf(&qmgr_ctx, argc, argv);
	SM_TEST(SM_SUCCESS == ret);
	if (sm_is_err(ret)) {
		exitcode = sm_is_perm_err(ret) ? EX_USAGE : EX_TEMPFAIL;
		goto error0;
	}
	qmgr_ctx.qmgr_cnf.q_cnf_iqdb_csize = iqdbsize;

#if QMGR_DEBUG
	for (u = 0; u < SM_ARRAY_SIZE(qm_debug); u++)
		qm_debug[u] = 99;
#endif

	qmgr_ctx.qmgr_tst_flags = Q_TST_NO_LI|Q_TST_NO_AR|Q_TST_NO_SCHED|Q_TST_NO_USR2|Q_TST_NO_QSSC;

	ret = qmgr_init1(&qmgr_ctx);
	SM_TEST(SM_SUCCESS == ret);
	if (sm_is_err(ret)) {
		sm_io_fprintf(smioerr, "sev=ERROR, func=main, qmgr_init1=%M\n", ret);
		goto error;
	}

	ret = qmgr_start(&qmgr_ctx);
	SM_TEST(SM_SUCCESS == ret);
	if (sm_is_err(ret)) {
		sm_io_fprintf(smioerr, "sev=ERROR, func=main, qmgr_start=%M\n", ret);
		goto error;
	}

	ret = t_iqdb(&qmgr_ctx, iqdbsize, nrcpts);
	SM_TEST(SM_SUCCESS == ret);
	if (sm_is_err(ret)) {
		sm_io_fprintf(smioerr, "sev=ERROR, func=main, t_iqdb=%M\n", ret);
		goto error;
	}

	(void) qm_info(&qmgr_ctx, smioout);
	iqdb_walk(qmgr_ctx.qmgr_iqdb, iqdb_print, (void *) NULL, THR_LOCK_UNLOCK);

	ret = qmgr_stop(&qmgr_ctx);
	SM_TEST(SM_SUCCESS == ret);
	if (sm_is_err(ret) && sm_error_value(ret) == SM_E_RSR_PRB)
		exitcode = EX_TEMPFAIL;
#if SM_HEAP_CHECK
	if (HEAP_CHECK)
		sm_heap_report(smioerr, 3);
#endif
	return sm_test_end();

  error:
	/* ignore shutdown errors... */
	(void) qmgr_stop(&qmgr_ctx);

  error0:
	/* select an appropriate error here... */
	return sm_test_end();
}
#else
int
main(int argc, char *argv[])
{
	sm_io_fprintf(smioerr, "SKIPPED %s: not compiled with QMGR_TEST\n", argv[0]);
	return 0;
}
#endif
